@@ -1,8 +1,10 @@
 use elrond_codec::TopEncode;
 
 use super::{BigIntApi, BigUintApi, ErrorApi, StorageReadApi, StorageWriteApi};
-use crate::hex_call_data::HexCallDataSerializer;
-use crate::types::{Address, ArgBuffer, AsyncCall, BoxedBytes, CodeMetadata, TokenIdentifier, Vec};
+use crate::{
+	types::{Address, ArgBuffer, AsyncCall, BoxedBytes, CodeMetadata, TokenIdentifier, Vec},
+	HexCallDataSerializer,
+};
 
 pub const ESDT_TRANSFER_STRING: &[u8] = b"ESDTTransfer";
 pub const ESDT_NFT_TRANSFER_STRING: &[u8] = b"ESDTNFTTransfer";
@@ -24,6 +26,10 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 	/// Same as the implementation from BlockchainApi.
 	fn get_sc_address(&self) -> Address;
 
+	/// To be used internally by the SendApi implementation.
+	/// Do not use directly from contracts. It might be removed from this trait at some point or reworked.
+	fn get_gas_left(&self) -> u64;
+
 	/// Sends EGLD to a given address, directly.
 	/// Used especially for sending EGLD to regular accounts.
 	fn direct_egld(&self, to: &Address, amount: &Self::AmountType, data: &[u8]);
@@ -38,25 +44,11 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 		arg_buffer: &ArgBuffer,
 	) -> Result<(), &'static [u8]>;
 
-	/// Sends an ESDT token to a given address, directly.
-	/// Used especially for sending ESDT to regular accounts.
-	///
-	/// Unlike sending ESDT via async call, this method can be called multiple times per transaction.
-	fn direct_esdt_via_transf_exec(
-		&self,
-		to: &Address,
-		token: &[u8],
-		amount: &Self::AmountType,
-		data: &[u8],
-	) -> Result<(), &'static [u8]> {
-		self.direct_esdt_execute(to, token, amount, 0, data, &ArgBuffer::new())
-	}
-
 	/// Sends ESDT to an address and executes like an async call, but without callback.
 	fn direct_esdt_execute(
 		&self,
 		to: &Address,
-		token: &[u8],
+		token: &TokenIdentifier,
 		amount: &Self::AmountType,
 		gas_limit: u64,
 		function: &[u8],
@@ -67,7 +59,7 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 	fn direct_esdt_nft_execute(
 		&self,
 		to: &Address,
-		token: &[u8],
+		token: &TokenIdentifier,
 		nonce: u64,
 		amount: &Self::AmountType,
 		gas_limit: u64,
@@ -75,6 +67,24 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 		arg_buffer: &ArgBuffer,
 	) -> Result<(), &'static [u8]>;
 
+	/// Performs a simple ESDT transfer, but via async call.
+	/// This is the preferred way to send ESDT.
+	fn transfer_esdt_via_async_call(
+		&self,
+		to: &Address,
+		token: &TokenIdentifier,
+		amount: &Self::AmountType,
+		data: &[u8],
+	) -> ! {
+		let mut serializer = HexCallDataSerializer::new(ESDT_TRANSFER_STRING);
+		serializer.push_argument_bytes(token.as_esdt_identifier());
+		serializer.push_argument_bytes(amount.to_bytes_be().as_slice());
+		if !data.is_empty() {
+			serializer.push_argument_bytes(data);
+		}
+		self.async_call_raw(&to, &Self::AmountType::zero(), serializer.as_slice())
+	}
+
 	/// Sends either EGLD or an ESDT token to the target address,
 	/// depending on what token identifier was specified.
 	fn direct(
@@ -87,68 +97,19 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 		if token.is_egld() {
 			self.direct_egld(to, amount, data);
 		} else {
-			let _ = self.direct_esdt_via_transf_exec(to, token.as_esdt_identifier(), amount, data);
+			let _ = self.direct_esdt_execute(to, token, amount, 0, data, &ArgBuffer::new());
 		}
 	}
 
-	/// Sends ESDT tokens to the target address. Handles any type of ESDT.
-	/// Note: this does not work with EGLD, use only with ESDT.
-	fn transfer_tokens(
+	fn direct_nft(
 		&self,
+		to: &Address,
 		token: &TokenIdentifier,
 		nonce: u64,
 		amount: &Self::AmountType,
-		to: &Address,
-	) {
-		if amount > &0 {
-			if nonce == 0 {
-				let _ =
-					self.direct_esdt_via_transf_exec(to, token.as_esdt_identifier(), amount, &[]);
-			} else {
-				let _ = self.direct_esdt_nft_via_transfer_exec(
-					to,
-					token.as_esdt_identifier(),
-					nonce,
-					amount,
-					&[],
-				);
-			}
-		}
-	}
-
-	/// Performs a simple ESDT transfer, but via async call.
-	/// This is the preferred way to send ESDT.
-	fn direct_esdt_via_async_call(
-		&self,
-		to: &Address,
-		esdt_token_name: &[u8],
-		amount: &Self::AmountType,
-		data: &[u8],
-	) -> ! {
-		let mut serializer = HexCallDataSerializer::new(ESDT_TRANSFER_STRING);
-		serializer.push_argument_bytes(esdt_token_name);
-		serializer.push_argument_bytes(amount.to_bytes_be().as_slice());
-		if !data.is_empty() {
-			serializer.push_argument_bytes(data);
-		}
-		self.async_call_raw(&to, &Self::AmountType::zero(), serializer.as_slice())
-	}
-
-	/// Sends either EGLD or an ESDT token to the target address,
-	/// depending on what token identifier was specified.
-	/// In case of ESDT it performs an async call.
-	fn direct_via_async_call(
-		&self,
-		to: &Address,
-		token: &TokenIdentifier,
-		amount: &Self::AmountType,
 		data: &[u8],
 	) {
-		if token.is_egld() {
-			let _ = self.direct_egld(to, amount, data);
-		} else {
-			self.direct_esdt_via_async_call(to, token.as_esdt_identifier(), amount, data);
-		}
+		let _ = self.direct_esdt_nft_execute(to, token, nonce, amount, 0, data, &ArgBuffer::new());
 	}
 
 	/// Sends an asynchronous call to another contract.
@@ -241,21 +202,21 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 	fn call_local_esdt_built_in_function(&self, gas: u64, function: &[u8], arg_buffer: &ArgBuffer);
 
 	/// Allows synchronous minting of ESDT tokens. Execution is resumed afterwards.
-	fn esdt_local_mint(&self, gas: u64, token: &[u8], amount: &Self::AmountType) {
+	fn esdt_local_mint(&self, token: &TokenIdentifier, amount: &Self::AmountType) {
 		let mut arg_buffer = ArgBuffer::new();
-		arg_buffer.push_argument_bytes(token);
+		arg_buffer.push_argument_bytes(token.as_esdt_identifier());
 		arg_buffer.push_argument_bytes(amount.to_bytes_be().as_slice());
 
-		self.call_local_esdt_built_in_function(gas, b"ESDTLocalMint", &arg_buffer);
+		self.call_local_esdt_built_in_function(self.get_gas_left(), b"ESDTLocalMint", &arg_buffer);
 	}
 
 	/// Allows synchronous burning of ESDT tokens. Execution is resumed afterwards.
-	fn esdt_local_burn(&self, gas: u64, token: &[u8], amount: &Self::AmountType) {
+	fn esdt_local_burn(&self, token: &TokenIdentifier, amount: &Self::AmountType) {
 		let mut arg_buffer = ArgBuffer::new();
-		arg_buffer.push_argument_bytes(token);
+		arg_buffer.push_argument_bytes(token.as_esdt_identifier());
 		arg_buffer.push_argument_bytes(amount.to_bytes_be().as_slice());
 
-		self.call_local_esdt_built_in_function(gas, b"ESDTLocalBurn", &arg_buffer);
+		self.call_local_esdt_built_in_function(self.get_gas_left(), b"ESDTLocalBurn", &arg_buffer);
 	}
 
 	/// Creates a new NFT token of a certain type (determined by `token_identifier`).  
@@ -263,8 +224,7 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 	/// This is a built-in function, so the smart contract execution is resumed after.
 	fn esdt_nft_create<T: elrond_codec::TopEncode>(
 		&self,
-		gas: u64,
-		token: &[u8],
+		token: &TokenIdentifier,
 		amount: &Self::AmountType,
 		name: &BoxedBytes,
 		royalties: &Self::AmountType,
@@ -273,7 +233,7 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 		uris: &[BoxedBytes],
 	) {
 		let mut arg_buffer = ArgBuffer::new();
-		arg_buffer.push_argument_bytes(token);
+		arg_buffer.push_argument_bytes(token.as_esdt_identifier());
 		arg_buffer.push_argument_bytes(amount.to_bytes_be().as_slice());
 		arg_buffer.push_argument_bytes(name.as_slice());
 		arg_buffer.push_argument_bytes(royalties.to_bytes_be().as_slice());
@@ -292,63 +252,54 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 			arg_buffer.push_argument_bytes(top_encoded_uri.as_slice());
 		}
 
-		self.call_local_esdt_built_in_function(gas, b"ESDTNFTCreate", &arg_buffer);
+		self.call_local_esdt_built_in_function(self.get_gas_left(), b"ESDTNFTCreate", &arg_buffer);
 	}
 
 	/// Adds quantity for an Non-Fungible Token. (which makes it a Semi-Fungible Token by definition)  
 	/// This is a built-in function, so the smart contract execution is resumed after.
-	fn esdt_nft_add_quantity(&self, gas: u64, token: &[u8], nonce: u64, amount: &Self::AmountType) {
-		let mut arg_buffer = ArgBuffer::new();
-		arg_buffer.push_argument_bytes(token);
-		arg_buffer.push_argument_bytes(&nonce.to_be_bytes()[..]);
-		arg_buffer.push_argument_bytes(amount.to_bytes_be().as_slice());
-
-		self.call_local_esdt_built_in_function(gas, b"ESDTNFTAddQuantity", &arg_buffer);
-	}
-
-	/// The reverse operation of `esdt_nft_add_quantity`, this locally decreases
-	/// This is a built-in function, so the smart contract execution is resumed after.
-	fn esdt_nft_burn(&self, gas: u64, token: &[u8], nonce: u64, amount: &Self::AmountType) {
-		let mut arg_buffer = ArgBuffer::new();
-		arg_buffer.push_argument_bytes(token);
-		arg_buffer.push_argument_bytes(&nonce.to_be_bytes()[..]);
-		arg_buffer.push_argument_bytes(amount.to_bytes_be().as_slice());
-
-		self.call_local_esdt_built_in_function(gas, b"ESDTNFTBurn", &arg_buffer);
-	}
-
-	/// Burns ESDT tokens. Handles any type of ESDT.
-	/// Note: this does not work with EGLD, use only with ESDT.
-	fn burn_tokens(
+	fn esdt_nft_add_quantity(
 		&self,
 		token: &TokenIdentifier,
 		nonce: u64,
 		amount: &Self::AmountType,
-		gas: u64,
 	) {
-		if amount > &0 {
-			if nonce == 0 {
-				self.esdt_local_burn(gas, token.as_esdt_identifier(), amount);
-			} else {
-				self.esdt_nft_burn(gas, token.as_esdt_identifier(), nonce, amount);
-			}
-		}
+		let mut arg_buffer = ArgBuffer::new();
+		arg_buffer.push_argument_bytes(token.as_esdt_identifier());
+		arg_buffer.push_argument_bytes(&nonce.to_be_bytes()[..]);
+		arg_buffer.push_argument_bytes(amount.to_bytes_be().as_slice());
+
+		self.call_local_esdt_built_in_function(
+			self.get_gas_left(),
+			b"ESDTNFTAddQuantity",
+			&arg_buffer,
+		);
+	}
+
+	/// The reverse operation of `esdt_nft_add_quantity`, this locally decreases
+	/// This is a built-in function, so the smart contract execution is resumed after.
+	fn esdt_nft_burn(&self, token: &TokenIdentifier, nonce: u64, amount: &Self::AmountType) {
+		let mut arg_buffer = ArgBuffer::new();
+		arg_buffer.push_argument_bytes(token.as_esdt_identifier());
+		arg_buffer.push_argument_bytes(&nonce.to_be_bytes()[..]);
+		arg_buffer.push_argument_bytes(amount.to_bytes_be().as_slice());
+
+		self.call_local_esdt_built_in_function(self.get_gas_left(), b"ESDTNFTBurn", &arg_buffer);
 	}
 
 	/// Performs a simple ESDT NFT transfer, but via async call.
 	/// This is the preferred way to send ESDT.
 	/// Note: call is done to the SC itself, so `from` should be the SCs own address
-	fn direct_esdt_nft_via_async_call(
+	fn transfer_esdt_nft_via_async_call(
 		&self,
 		from: &Address,
 		to: &Address,
-		token: &[u8],
+		token: &TokenIdentifier,
 		nonce: u64,
 		amount: &Self::AmountType,
 		data: &[u8],
 	) {
 		let mut serializer = HexCallDataSerializer::new(ESDT_NFT_TRANSFER_STRING);
-		serializer.push_argument_bytes(token);
+		serializer.push_argument_bytes(token.as_esdt_identifier());
 		serializer.push_argument_bytes(&nonce.to_be_bytes()[..]);
 		serializer.push_argument_bytes(amount.to_bytes_be().as_slice());
 		serializer.push_argument_bytes(to.as_bytes());
@@ -357,19 +308,4 @@ pub trait SendApi: ErrorApi + Clone + Sized {
 		}
 		self.async_call_raw(&from, &Self::AmountType::zero(), serializer.as_slice());
 	}
-
-	/// Sends an ESDT NFT to a given address, directly.
-	/// Used especially for sending ESDT to regular accounts.
-	///
-	/// Unlike sending ESDT via async call, this method can be called multiple times per transaction.
-	fn direct_esdt_nft_via_transfer_exec(
-		&self,
-		to: &Address,
-		token: &[u8],
-		nonce: u64,
-		amount: &Self::AmountType,
-		data: &[u8],
-	) -> Result<(), &'static [u8]> {
-		self.direct_esdt_nft_execute(to, token, nonce, amount, 0, data, &ArgBuffer::new())
-	}
 }
